/*

	This is a hacked up driver for the SiS630 ACPI MSG LED
	Written by John Newbigin <jn@it.swin.edu.au>

	This module is based on a module from the lm_sensors package and is
	covered by the GPL.

	It can set the LED to on, off or 1Hz flash

*/

#include <linux/version.h>
#include <linux/pci.h>
#include <asm/io.h>
#include <linux/kernel.h>
#include <linux/stddef.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/sysctl.h>
#include <asm/errno.h>
#include <linux/types.h>
#include <linux/init.h>


#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,2,18)) || \
    (LINUX_VERSION_CODE == KERNEL_VERSION(2,3,0))
#define init_MUTEX(s) do { *(s) = MUTEX; } while(0)
#endif

#ifndef THIS_MODULE
#define THIS_MODULE NULL
#endif

void set_led_state(int state);

static ssize_t module_input(
    struct file *file,   /* The file itself */
    const char *buf,     /* The buffer with input */
    size_t length,       /* The buffer's length */
    loff_t *offset)      /* offset to file - ignore */
{
	int state;

	if(current->euid == 0)
	{
		if(length > 0)
		{
			state = buf[0] - '0';
			if(state >= 0 && state <= 2)
			{
	   			printk(KERN_INFO "LED: change state to %d\n", state);
				set_led_state(state);
			}
			else
			{
	   			printk(KERN_INFO "LED: invalid state\n");
			}
		}

		return length;
	}
	else
	{
		return -EACCES;
	}
}


int module_open(struct inode *inode, struct file *file)
{
  MOD_INC_USE_COUNT;
 
  return 0;
}

int module_close(struct inode *inode, struct file *file)
{
  MOD_DEC_USE_COUNT;

  return 0; 
}

static struct file_operations File_Ops_4_Our_Proc_File =
  {
    NULL, 		// module_owner
    NULL, 		// lseek 
    NULL, 		// read
    NULL, 		// new read
    module_input,	// write
    NULL, 		// new write
    NULL, 		// readdir
    NULL, 		// poll
    NULL, 		// ioctl
    NULL, 		// mmap
    module_open,	// open
    NULL,   		// flush
    module_close,	//close
  };


EXPORT_NO_SYMBOLS;

#ifdef MODULE

MODULE_AUTHOR("John Newbigin <jn@it.swin.edu.au>");
MODULE_DESCRIPTION("SiS630 ACPI MSG LED device");
MODULE_LICENSE("GPL");

/*
  state 0 = off
  state 1 = on
  state 2 = flash

*/

void set_led_state(int state)
{
	int flags;
	flags = inw(0xe462);

	// clear the LED flags
	flags &= (~0xC0);
	// led is now on.

	if(state == 2)
	{
		// we can flash
		flags |= 0x40;
	}
	else if(state == 0)
	{
		// of turn off
		flags |= 0x80;
	}

	printk(KERN_INFO "LED: new flags = 0x%08X\n", flags);
	outw(flags, 0xe462);
}

int init_module(void)
{
	struct proc_dir_entry   *led_proc;

	if(request_region(0xe462, 0x02, "SiS630 ACPI MSG LED"))
	{
		led_proc = create_proc_entry("msg_led", S_IWUSR, NULL);
		if(led_proc)
		{
			led_proc->proc_fops = &File_Ops_4_Our_Proc_File;
		}
		else
		{
			release_region(0xe462, 0x02);
			return 1;
		}
		return 0;
	}
	else
	{
		return 1;
	}
}

int cleanup_module(void)
{
	release_region(0xe462, 0x02);
	return 0;
}

#endif				/* MODULE */
